"""Test configuration."""
import os

import pytest


@pytest.fixture()
def dont_run_in_appveyor():
    """Test if the app is not on AppVeyor."""
    if is_running_in_appveyor():
        pytest.skip("This test does not work on appveyor")

    return True


def is_running_in_appveyor():
    """Check if the AppVeyor environment variable is set."""
    return os.environ.get("APPVEYOR", "").lower() == "true"



"""FSM Descriptors."""
from typing import Callable, List, Optional

TucoCallback = Callable[[object], bool]


class BaseState:
    """Base state."""

    pass


class State(BaseState):
    """Describe the flow of a state."""

    def __init__(
        self,
        events: Optional[List["Event"]] = None,
        error: Optional["Error"] = None,
        timeout: Optional["Timeout"] = None,
        on_enter: Optional[List[TucoCallback]] = None,
    ) -> None:
        """Initialize default values."""
        self.events = events or []  # type: List[Event]
        self.error = error
        self.timeout = timeout
        self.on_enter = on_enter or []  # type: List[TucoCallback]


class FinalState(BaseState):
    """Dumb class to mark final states."""

    def __init__(self, on_enter: Optional[List[TucoCallback]] = None) -> None:
        """Initialize default values."""
        self.timeout = None
        self.on_enter = on_enter or []  # type: List[TucoCallback]


class Event:
    """Describe an event."""

    def __init__(self, event_name, target_state, commands=None, error=None) -> None:
        """Initialize default values."""
        self.event_name = event_name
        self.target_state = target_state
        self.commands = commands or []
        self.error = error

    def __repr__(self) -> str:
        """Basic representation."""
        return "<FSM Event {!r} with target state {!r}>".format(self.event_name, self.target_state)


class Error:
    """Error handling."""

    def __init__(self, target_state, commands=None) -> None:
        """Initialize default values."""
        self.target_state = target_state
        self.commands = commands or []


class Timeout:
    """Timeout class."""

    def __init__(self, timedelta, target_state, commands=None) -> None:
        """Initialize default values."""
        self.timedelta = timedelta
        self.target_state = target_state
        self.commands = commands or []



"""Basic lock interface."""
from tuco.exceptions import TucoDoNotLockError


class BaseLock:
    """Common lock functions."""

    def __init__(self, fsm, id_field):
        """Hold the fsm and it's id field."""
        self.fsm = fsm
        self.id_field = id_field

    @property
    def hash_key(self) -> str:
        """Generate a hash key to be used when locking an object."""
        primary_key = getattr(self.fsm.container_object, self.id_field, None)
        if primary_key:
            return "fsm_{}_pk_{}".format(self.fsm.__class__.__name__, primary_key)

        raise TucoDoNotLockError()

    def lock(self) -> bool:
        """Implement lock."""
        raise NotImplementedError()

    def unlock(self) -> bool:
        """Implement unlock."""
        raise NotImplementedError()



"""FSM Decorators."""
from functools import wraps


def on_change(original_function):
    """Register on change event on the state machine."""

    @wraps(original_function)
    def decorated(*args, **kwargs):
        """Just run the event."""
        return original_function(*args, **kwargs)

    setattr(decorated, "_on_change_event", True)
    return decorated


def on_error(original_function):
    """Register on error event on the state machine."""

    @wraps(original_function)
    def decorated(*args, **kwargs):
        """Just run the event."""
        return original_function(*args, **kwargs)

    setattr(decorated, "_on_error_event", True)
    return decorated



"""FSM Base tests."""

import os
import threading
from datetime import timedelta
from unittest import mock

import pytest
import pytz

from tests.example_fsm import ExampleCreditCardFSM, StateHolder
from tuco import FSM, properties
from tuco.decorators import on_change, on_error
from tuco.exceptions import TucoAlreadyLockedError, TucoEventNotFoundError, TucoInvalidStateChangeError
from tuco.locks import RedisLock


def test_state_changing():
    """Test basic changing."""
    fsm = ExampleCreditCardFSM(StateHolder())
    assert fsm.container_object.current_state == "new"

    assert fsm.event_allowed("Initialize")
    assert fsm.trigger("Initialize")
    assert fsm.current_state == "authorisation_pending"

    assert fsm.event_allowed("Authorize")
    assert fsm.trigger("Authorize")
    assert fsm.current_state == "capture_pending"

    assert fsm.event_allowed("Capture")
    assert fsm.trigger("Capture")
    assert fsm.current_state == "paid"

    assert fsm.event_allowed("Initialize") is False
    with pytest.raises(TucoEventNotFoundError):
        assert fsm.trigger("Initialize")
    assert fsm.container_object.current_state == "paid"


def test_command_execution():
    """Test commands."""
    command1, command2 = mock.Mock(), mock.Mock()
    state_holder = StateHolder()

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "new"
        state_attribute = "current_state"

        new = properties.State(events=[properties.Event("TestEvent", "final_state", commands=[command1, command2])])
        final_state = properties.FinalState()

    fsm = TestFSM(state_holder)
    fsm.trigger("TestEvent")

    command1.assert_called_once_with(state_holder)
    command2.assert_called_once_with(state_holder)
    assert fsm.current_state == "final_state"


def test_command_with_error():
    """Test event behavior when a command fail."""
    command1 = mock.Mock()
    command2 = mock.Mock(side_effect=RuntimeError)
    command3 = mock.Mock()

    state_holder = StateHolder()

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "new"
        state_attribute = "current_state"

        new = properties.State(
            events=[properties.Event("TestEvent", "final_state", commands=[command1, command2, command3])]
        )
        final_state = properties.FinalState()

    fsm = TestFSM(state_holder)
    with pytest.raises(RuntimeError):
        fsm.trigger("TestEvent")

    command1.assert_called_once_with(state_holder)
    command2.assert_called_once_with(state_holder)
    assert command3.call_count == 0
    assert fsm.current_state == "new"


def test_invalid_error():
    """Test common cases of invalid error configured."""
    with pytest.raises(RuntimeError):

        class TestFSM(FSM):
            """Dumb class."""

            state = properties.State(error=properties.Error("invalid_state"))

        assert TestFSM

    with pytest.raises(RuntimeError):

        class TestFSM2(FSM):
            """Dumb class."""

            state = properties.State(
                events=[properties.Event("EventName", "final_state", error=properties.Error("invalid_state"))]
            )
            final_state = properties.FinalState()

        assert TestFSM2


def test_errors():
    """Test errors."""
    event_error_command = mock.Mock()
    state_error_command = mock.Mock()

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "state"

        state = properties.State(
            error=properties.Error("state_error", commands=[state_error_command]),
            events=[
                properties.Event(
                    "TriggerEventError",
                    "non_reachable",
                    error=properties.Error("event_error", commands=[event_error_command]),
                    commands=[lambda x: False],
                ),
                properties.Event("TriggerStateError", "non_reachable", commands=[lambda x: False]),
            ],
        )

        state_error = properties.FinalState()
        event_error = properties.FinalState()
        non_reachable = properties.FinalState()

    fsm = TestFSM(StateHolder())
    assert fsm.current_state == "state"
    assert fsm.trigger("TriggerEventError") is False
    assert fsm.current_state == "event_error"
    event_error_command.assert_called_once_with(fsm.container_object)

    fsm = TestFSM(StateHolder())
    assert fsm.current_state == "state"
    assert fsm.trigger("TriggerStateError") is False
    assert fsm.current_state == "state_error"
    state_error_command.assert_called_once_with(fsm.container_object)


def test_invalid_state_changes():
    """Test cases when the current state is being change to an invalid one."""

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "state1"

        state1 = properties.State(events=[properties.Event("ChangeState", target_state="state2")])

        state2 = properties.FinalState()

    fsm = TestFSM(StateHolder())
    assert fsm.current_state == "state1"

    fsm.current_state = "state2"
    with pytest.raises(TucoInvalidStateChangeError):
        fsm.current_state = "state1"

    assert fsm.current_state == "state2"

    class TestFSM2(FSM):
        """Dumb class."""

        initial_state = "state1"

        state1 = properties.State(events=[properties.Event("ChangeState", target_state="state2")])

        state2 = properties.State(events=[properties.Event("ChangeState", target_state="state3")])
        state3 = properties.FinalState()

    fsm2 = TestFSM2(StateHolder())
    assert fsm2.current_state == "state1"

    fsm2.current_state = "state2"
    with pytest.raises(TucoInvalidStateChangeError):
        fsm2.current_state = "state1"

    assert fsm2.current_state == "state2"


def test_timeout():
    """Test timeout behaviors."""
    command = mock.Mock()

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "new"

        new = properties.State(events=[properties.Event("state1", "state1")])

        state1 = properties.State(timeout=properties.Timeout(timedelta(days=7), "timeout", commands=[command]))

        timeout = properties.FinalState()

        @property
        def current_state_date(self):
            """Database always send time zone aware dates but not in tests."""
            return getattr(self.container_object, self.date_attribute).replace(tzinfo=pytz.UTC)

    fsm = TestFSM(StateHolder())
    fsm.trigger("state1")
    assert fsm.trigger_timeout() is False
    assert fsm.current_state == "state1"
    assert command.call_count == 0

    fsm.container_object.current_state_date -= timedelta(days=7)
    assert fsm.trigger_timeout() is True
    assert fsm.current_state == "timeout"
    command.assert_called_once_with(fsm.container_object)


def test_list_timeouts():
    """Test the ability to list all configured timeouts on state machine. Useful to query expired objects."""

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "new"

        new = properties.State(events=[properties.Event("state1", "state1")])

        state1 = properties.State(timeout=properties.Timeout(timedelta(days=6), "timeout"))
        state2 = properties.State(timeout=properties.Timeout(timedelta(days=5), "timeout"))

        timeout = properties.FinalState()

    timeouts = list(TestFSM.get_all_timeouts())
    assert len(timeouts) == 2


def test_on_enter():
    """Test on enter event."""
    command = mock.Mock()

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "state1"

        state1 = properties.State(on_enter=[command])

    fsm = TestFSM(StateHolder())
    assert fsm.current_state == "state1"
    command.assert_called_once_with(fsm.container_object)


def test_on_change():
    """Test on change."""
    command = mock.Mock()

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "state1"

        @on_change
        def hacky_change_call(self, *args, **kwargs):
            """Hacky way to check on change calls."""
            command(self, *args, **kwargs)

        state1 = properties.State(events=[properties.Event("Change", "final_state")])
        final_state = properties.FinalState()

    fsm = TestFSM(StateHolder())
    assert command.call_count == 0

    command.reset_mock()

    assert fsm.trigger("Change")

    assert command.call_count == 1
    (arg1, arg2, arg3) = command.call_args[0]  # pylint: disable=unsubscriptable-object
    assert arg1 == fsm
    assert arg2.current_state == "state1"
    assert arg3.current_state == "final_state"


def test_on_error():
    """Test on error."""
    command = mock.Mock()
    throw_exception = mock.Mock(side_effect=NotADirectoryError)

    class TestFSM(FSM):
        """Dumb class."""

        initial_state = "state1"

        @on_error
        def hacky_error_call(self, *args, **kwargs):
            """Hacky way to check on error calls."""
            command(self, *args, **kwargs)

        state1 = properties.State(events=[properties.Event("Change", "final_state", commands=[throw_exception])])
        final_state = properties.FinalState()

    fsm = TestFSM(StateHolder())
    with pytest.raises(NotADirectoryError):
        fsm.trigger("Change")
    assert command.call_count == 1
    assert len(command.call_args[0]) == 4  # pylint: disable=unsubscriptable-object
    (self, old_state, new_state, exception) = command.call_args[0]  # pylint: disable=unsubscriptable-object
    assert self == fsm
    assert old_state == "state1"
    assert new_state == "final_state"
    assert isinstance(exception, NotADirectoryError)


def test_locking():
    """Testing locking system."""
    hold_triggered = threading.Event()
    shutdown = threading.Event()

    class TestFSM(FSM):
        """Dumb class."""

        # pylint: disable=no-self-use
        def hold(self):
            """Hold the state machine."""
            hold_triggered.set()
            shutdown.wait()

        new = properties.State(events=[properties.Event("Hold", "final_state", commands=[hold])])
        final_state = properties.FinalState()

    def lock_it_all():
        """Lock the state machine to test it."""
        with TestFSM(StateHolder()) as fsm:
            fsm.trigger("Hold")

    worker = threading.Thread(target=lock_it_all, daemon=True)
    worker.start()

    # Hold until the machine is really locked.
    hold_triggered.wait(timeout=1)

    with pytest.raises(TucoAlreadyLockedError):
        with TestFSM(StateHolder()):
            pass

    shutdown.set()
    worker.join()


def test_locking_without_id():
    """Make sure that items without id won't get locked."""
    hold_triggered = threading.Event()
    shutdown = threading.Event()

    class TestFSM(FSM):
        """Dumb class."""

        # pylint: disable=no-self-use
        def hold(self):
            """Hold the state machine."""
            hold_triggered.set()
            shutdown.wait()

        new = properties.State(
            events=[properties.Event("Hold", "final_state", commands=[hold]), properties.Event("Finish", "final_state")]
        )
        final_state = properties.FinalState()

    state_holder = StateHolder()
    state_holder.id = None

    def lock_it_all():
        """Lock the state machine to test it."""
        with TestFSM(state_holder) as fsm:
            fsm.trigger("Hold")

    worker = threading.Thread(target=lock_it_all, daemon=True)
    worker.start()

    # Hold until the machine is really locked.
    hold_triggered.wait()

    with TestFSM(state_holder) as fsm:
        fsm.trigger("Finish")
        assert fsm.current_state == "final_state"

    shutdown.set()
    worker.join()


def test_redis_locking(dont_run_in_appveyor):
    """Testing redis locking system."""
    assert dont_run_in_appveyor  # After we install redis in appveyor we can remove this
    hold_triggered = threading.Event()
    shutdown = threading.Event()

    class ConfiguredRedisLock(RedisLock):
        def __init__(self, *args, **kwargs):
            import redis

            os.environ.setdefault("REDIS_SERVER", "127.0.0.1")
            super().__init__(10, redis.StrictRedis(os.environ["REDIS_SERVER"]), *args, **kwargs)

    class TestFSM(FSM):
        """Dumb class."""

        lock_class = ConfiguredRedisLock

        new = properties.FinalState()

    def lock_it_all():
        """Lock the state machine to test it."""
        with TestFSM(StateHolder()):
            hold_triggered.set()
            shutdown.wait()

    worker = threading.Thread(target=lock_it_all, daemon=True)
    worker.start()

    # Hold until the machine is really locked.
    hold_triggered.wait(timeout=1.5)  # It it timeout it means that the thread broke somehow

    with pytest.raises(TucoAlreadyLockedError):
        with TestFSM(StateHolder()):
            pass

    shutdown.set()
    worker.join()


def test_fatal_error():
    """Test if we can change current state to fatal state."""
    fsm = ExampleCreditCardFSM(StateHolder())
    assert fsm.current_state == "new"
    fsm.current_state = fsm.fatal_state
    assert fsm.current_state == fsm.fatal_state



"""Meta class to validate FSM implementations on parsing time."""
import collections

from tuco.properties import BaseState, Event, FinalState


class FSMBase(type):
    """Metaclass to generate all FSM models."""

    def __new__(mcs, name, bases, attributes):
        """Make sure that we don't have broken state machines and hide some stuff from users."""
        super_new = super(FSMBase, mcs).__new__

        # Propagate existing __classcell__ to fix a DeprecationWarning: __class__ not set defining
        # 'MyClass' as <class 'path.to.MyClass'>. Was __classcell__ propagated to type.__new__?
        module = attributes.pop("__module__", None)
        new_namespace = {"__module__": module}
        if "__classcell__" in attributes:
            new_namespace["__classcell__"] = attributes["__classcell__"]

        new_class = super_new(mcs, name, bases, new_namespace)
        for name, value in attributes.items():
            if isinstance(value, BaseState):
                mcs._create_state(new_class, name, value)
            else:
                setattr(new_class, name, value)
                if callable(value):
                    for event_name in ("_on_change_event", "_on_error_event"):
                        if getattr(value, event_name, False):
                            setattr(new_class, event_name, value)

        mcs._validate_timeouts(new_class)
        mcs._validate_errors(new_class)
        mcs._validate_events(new_class)
        return new_class

    @staticmethod
    def _create_state(new_class, name, value) -> None:
        """Create a state hidden in new class."""
        if new_class._states is None:
            new_class._states = {}

        states = new_class._states
        if name in states:
            raise RuntimeError("Duplicated states are not allowed {!r} {!r}".format(new_class, name))

        states[name] = value

    @staticmethod
    def _validate_timeouts(new_class) -> None:
        """Validate all timeouts."""
        states = new_class._states
        if states is None:
            return
        for name, state in states.items():
            if isinstance(state, FinalState):
                continue
            if state.timeout and state.timeout.target_state not in states:
                raise RuntimeError(
                    "Invalid timeout target state, {!r} not found in {!r}.".format(
                        state.timeout.target_state, new_class
                    )
                )
            if name == new_class.initial_state and state.timeout:
                raise RuntimeError("Initial state cannot have a timeout {!r} {!r}.".format(new_class, state))

    @staticmethod
    def _validate_events(new_class) -> None:
        """Make a pre-validation on events to fail fast."""
        states = new_class._states
        if states is None:
            return
        for parent_state_name, parent_state in states.items():
            if isinstance(parent_state, FinalState):
                continue

            counter = collections.Counter([event.event_name for event in parent_state.events])
            duplicated_events = [(element, amount) for element, amount in counter.items() if amount > 1]
            if duplicated_events:
                raise RuntimeError(
                    "Duplicated events found in state {} {!r} {}".format(
                        parent_state_name, parent_state, duplicated_events
                    )
                )

            for event in parent_state.events:
                FSMBase._validate_event(event, parent_state_name, parent_state, new_class)

    @staticmethod
    def _validate_event(event, parent_state_name, parent_state, new_class) -> None:
        """Validate an event."""
        if not isinstance(event, Event):
            raise RuntimeError("Invalid event class when parsing state {} in {!r}".format(parent_state_name, new_class))

        if event.target_state not in getattr(new_class, "_states", {}):
            raise RuntimeError(
                "Target state not found {} in {!r} ({})".format(event.target_state, parent_state, parent_state_name)
            )

    @staticmethod
    def _validate_errors(new_class) -> None:
        """Walk through states and events to check if events are not going to dead ends."""
        states = new_class._states
        if states is None:
            return
        for _, state in states.items():
            if isinstance(state, FinalState):
                continue

            if state.error:
                FSMBase._validate_error(new_class, state.error)

            for event in state.events:
                if event.error:
                    FSMBase._validate_error(new_class, event.error)

    @staticmethod
    def _validate_error(new_class, error) -> None:
        """Check if a declared error is properly configured."""
        states = new_class._states
        if not states or error.target_state not in states:
            raise RuntimeError("Could not find target state {} inside {!r}".format(error.target_state, new_class))



"""Graph ceeation tests."""
import os
from xml.etree.ElementTree import fromstring

import pytest

from tests.example_fsm import ExampleCreditCardFSM
from tuco import FSM
from tuco.exceptions import TucoEmptyFSMError


def test_svg(dont_run_in_appveyor):
    """Test SVG generation.

    It is hard to test SVGs generated from different graphviz version so we just do a parsing and assert number of
    elements.
    """
    assert dont_run_in_appveyor
    with open(os.path.join(os.path.dirname(__file__), "fixtures", "sample-fsm.svg")) as f:
        generated_graph = fromstring(ExampleCreditCardFSM.generate_graph("svg"))
        saved_graph = fromstring(f.read())
        assert len(list(generated_graph[0])) == len(list(saved_graph[0]))


def test_empty_fsm(dont_run_in_appveyor):
    """Test a case where FSM has no state and tries to generate a graph."""
    assert dont_run_in_appveyor
    new_class = type("EmptyFSM", (FSM,), {})  # type: FSM
    with pytest.raises(TucoEmptyFSMError):
        new_class.generate_graph()



from typing import Optional  # noqa

from tuco.exceptions import TucoAlreadyLockedError, TucoDoNotLockError

from .base import BaseLock


class RedisLock(BaseLock):
    """Redis lock.

    This class should be extended and provided with timeout and a redis connection pool as there is no way to know
    where the user stores this data, it can be inside some config or a flask app for example.
    """

    def __init__(self, lock_timeout, redis_connection, *args, **kwargs) -> None:
        """Start the lock with default timeout."""
        super().__init__(*args, **kwargs)
        from redis.lock import Lock  # noqa

        self._lock = None  # type: Optional[Lock]
        self.lock_timeout = lock_timeout
        self.redis_connection = redis_connection

    def lock(self) -> bool:
        """Lock an object."""
        try:
            hash_key = self.hash_key
        except TucoDoNotLockError:
            return True

        self._lock = self.redis_connection.lock(hash_key, timeout=self.lock_timeout)

        if not self._lock.acquire(False):
            raise TucoAlreadyLockedError()
        return True

    def unlock(self) -> bool:
        """Unlock an object."""
        from redis.exceptions import LockError

        if self._lock is None:
            return True
        try:
            self._lock.release()
        except LockError:
            pass

        return True



"""Base classes to be used in FSM."""
import copy
from datetime import datetime
from typing import Dict, Iterator, List, Tuple, Type  # noqa

import pytz

from tuco.exceptions import (
    TucoAlreadyLockedError,
    TucoEventNotFoundError,
    TucoInvalidStateChangeError,
    TucoInvalidStateHolderError,
)
from tuco.locks import MemoryLock
from tuco.locks.base import BaseLock  # noqa
from tuco.meta import FSMBase
from tuco.properties import Event, FinalState, State, Timeout

__all__ = ("FSM",)

mockable_utcnow = datetime.utcnow  # Easier to write tests


class FSM(metaclass=FSMBase):
    """Class that handle event transitions.

    Your state machines should extend from this.
    """

    #: The default initial state is "new" but can be overridden
    initial_state = "new"
    state_attribute = "current_state"
    date_attribute = "current_state_date"
    id_field = "id"

    fatal_state = "fatal_error"

    lock_class = MemoryLock  # type: Type[BaseLock]
    _states = None  # type: Dict[str, State]

    def __init__(self, container_object) -> None:
        """Initialize the container object with the initial state."""
        self.container_object = container_object
        for field in (self.state_attribute, self.date_attribute, self.id_field):
            if not hasattr(container_object, field):
                raise TucoInvalidStateHolderError(
                    "Required field {!r} not found inside {!r}.".format(field, container_object)
                )
        if self.current_state is None:
            self.current_state = self.initial_state

        self.lock = self.lock_class(self, self.id_field)

    def __enter__(self) -> "FSM":
        """Lock the state machine."""
        self.lock.lock()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """If TucoAlreadyLockedError did not throw, unlock the machine."""
        if exc_type and issubclass(exc_type, TucoAlreadyLockedError):
            return

        self.lock.unlock()

    def __repr__(self) -> str:
        """Basic representation."""
        return "<{} - current_state {!r} with holder {} - ID {!r}>".format(
            self.__class__.__name__,
            self.current_state,
            self.container_object.__class__.__name__,
            getattr(self.container_object, self.id_field),
        )

    @property
    def current_time(self) -> datetime:
        """Return utcnow and should be extended if you care about time zone or something else."""
        return mockable_utcnow()

    @property
    def current_state_date(self) -> datetime:
        """Return current date stored in object."""
        return getattr(self.container_object, self.date_attribute)

    @property
    def current_state(self) -> str:
        """Return the current state stored in object."""
        return getattr(self.container_object, self.state_attribute)

    @current_state.setter
    def current_state(self, new_state) -> None:
        """Set a state on container object."""
        call_on_change = bool(self.current_state)
        old_state = copy.copy(self.container_object)
        if new_state != self.fatal_state:
            if not self.state_allowed(new_state):
                raise TucoInvalidStateChangeError(
                    "Old state {!r}, new state {!r}.".format(self.current_state, new_state)
                )

            setattr(self.container_object, self.state_attribute, new_state)
            setattr(self.container_object, self.date_attribute, self.current_time)
            for command in self.current_state_instance.on_enter:
                command(self.container_object)
        else:
            setattr(self.container_object, self.state_attribute, new_state)
            setattr(self.container_object, self.date_attribute, self.current_time)

        if call_on_change:
            self._call_on_change(old_state, self.container_object)

    def state_allowed(self, state_name) -> bool:
        """Check if the transition to the new state is allowed."""
        if self.current_state is None and state_name == self.initial_state:
            return True

        if isinstance(self.current_state_instance, FinalState):
            return False

        if self.current_state_instance.timeout and self.current_state_instance.timeout.target_state == state_name:
            return True

        if any(event for event in self.possible_events if event.target_state == state_name):
            return True

        current_state = self.current_state_instance
        if current_state.error and current_state.error.target_state == state_name:
            return True

        for event in current_state.events:
            if event.error and event.error.target_state == state_name:
                return True

        return False

    @property
    def current_state_instance(self) -> State:
        """Return the current `State` instance."""
        return self._states[self.current_state]

    @property
    def possible_events(self) -> List[Event]:
        """Return all possible events for the current state."""
        return self.possible_events_from_state(self.current_state)

    @classmethod
    def possible_events_from_state(cls, state_name) -> List[Event]:
        """Return all possible events from a specific state.

        :param state_name: State to check
        """
        state = cls._states[state_name]
        return getattr(state, "events", [])

    def _get_event(self, event_name) -> Event:
        """Get an event inside current state based on it's name."""
        for event in self.possible_events:
            if event.event_name == event_name:
                return event

        raise TucoEventNotFoundError(
            "Event {!r} not found in {!r} on current state {!r}".format(
                event_name, [event.event_name for event in self.possible_events], self.current_state
            )
        )

    def event_allowed(self, event_name) -> bool:
        """Check if is possible to run an event.

        :param event_name: Event to check.
        """
        try:
            self._get_event(event_name)
        except TucoEventNotFoundError:
            return False

        return True

    def _trigger_error(self, event) -> None:
        """Search for an error handler inside event, and then inside state."""
        if event.error:
            error = event.error
        else:
            error = self._states[self.current_state].error
            if not error:
                return

        for command in error.commands:
            command(self.container_object)

        self.current_state = error.target_state

    def trigger(self, event_name, *args, **kwargs) -> bool:
        """Trigger an event and call its commands with specified arguments..

        :param event_name: Event to execute.
        """
        event = self._get_event(event_name)
        for command in event.commands:
            try:
                return_value = command(self.container_object, *args, **kwargs)
            except Exception as e:
                self._call_on_error(e, event.target_state)
                raise

            if not return_value:
                self._trigger_error(event)
                return False

        self.current_state = event.target_state

        return True

    def trigger_timeout(self) -> bool:
        """Trigger timeout if it's possible."""
        timeout = self.current_state_instance.timeout

        if not timeout:
            return False

        if datetime.utcnow().replace(tzinfo=pytz.UTC) < (self.current_state_date + timeout.timedelta):
            return False

        for command in timeout.commands:
            try:
                command(self.container_object)
            except Exception as e:
                self._call_on_error(e, timeout.target_state)
                raise

        self.current_state = timeout.target_state
        return True

    @classmethod
    def get_all_states(cls) -> Dict[str, State]:
        """List all states for this state machine."""
        return cls._states

    @classmethod
    def get_all_timeouts(cls) -> Iterator[Tuple[str, Timeout]]:
        """List all configured timeouts for this state machine."""
        for state_name, state in cls._states.items():
            if isinstance(state, FinalState) or not state.timeout:
                continue
            yield (state_name, state.timeout)

    @classmethod
    def get_all_finals(cls) -> Iterator[FinalState]:
        """List all configured final states for this state machine."""
        for state_name, state in cls._states.items():
            if isinstance(state, FinalState):
                yield state_name

    def _call_on_change(self, old_state, new_state) -> None:
        """If on_change function exists, call it.

        :param old_state: A shallow copy of the holder object.
        :param new_state: The changed version of the object holder.
        """
        function = getattr(self, "_on_change_event", None)
        if function:
            function(old_state, new_state)

    def _call_on_error(self, exception, new_state) -> None:
        """If on_error function exists, call it."""
        function = getattr(self, "_on_error_event", None)
        if function:
            function(self.current_state, new_state, exception)

    @classmethod
    def generate_graph(cls, file_format="svg") -> str:
        """Generate a SVG graph."""
        from .graph_builder import generate_from_class

        return generate_from_class(cls, file_format)



"""Examples FSM for tests."""
from datetime import timedelta

from tuco import FSM, properties


class StateHolder:
    """A simple class to hold current states."""

    def __init__(self):
        """Just initialize with None."""
        self.current_state = None
        self.current_state_date = None
        self.id = 1234


class ExampleCreditCardFSM(FSM):
    """Credit card FSM."""

    initial_state = "new"
    state_attribute = "current_state"

    event_error = properties.FinalState()
    state_error = properties.FinalState()

    new = properties.State(
        events=[properties.Event("Initialize", "authorisation_pending", error=properties.Error("event_error"))],
        error=properties.Error("state_error"),
    )

    authorisation_pending = properties.State(events=[properties.Event("Authorize", "capture_pending")])

    capture_pending = properties.State(
        events=[properties.Event("Capture", "paid")], timeout=properties.Timeout(timedelta(days=7), "timeout_test")
    )

    timeout_test = properties.FinalState()

    paid = properties.State(events=[properties.Event("Refund", "refund_pending")])

    finished = properties.State(
        events=[properties.Event("ChargeBack", "charged_back"), properties.Event("Refund", "refunded")]
    )

    refund_pending = properties.State(events=[properties.Event("Refund", "refunded")])

    refunded = properties.FinalState()
    charged_back = properties.FinalState()



"""FSM Exceptions."""


class TucoException(Exception):
    """Base FSM Exception."""

    pass


class TucoEventNotFoundError(TucoException):
    """Event not found."""

    pass


class TucoInvalidStateChangeError(TucoException):
    """In case someone try to change to an invalid state."""

    pass


class TucoInvalidStateHolderError(TucoException):
    """In case the state holder do not have all required attributes."""

    pass


class TucoDoNotLockError(TucoException):
    """In case we don't have and ID, means that the model is new."""

    pass


class TucoAlreadyLockedError(TucoException):
    """When an object is already locked."""

    pass


class TucoEmptyFSMError(TucoException):
    """FSM has no state defined."""

    pass



"""Memory lock module."""
from threading import RLock
from typing import Dict  # noqa

from tuco.exceptions import TucoAlreadyLockedError, TucoDoNotLockError

from .base import BaseLock


class MemoryLock(BaseLock):
    """Simple memory lock implementation."""

    global_lock = RLock()
    locks = {}  # type: Dict[str, str]

    def lock(self) -> bool:
        """Lock an object."""
        try:
            hash_key = self.hash_key
        except TucoDoNotLockError:
            return True

        with self.global_lock:
            if hash_key in self.locks:
                raise TucoAlreadyLockedError()
            self.locks[hash_key] = "locked"
            return True

    def unlock(self) -> bool:
        """Unlock an object."""
        try:
            hash_key = self.hash_key
        except TucoDoNotLockError:
            return True

        with self.global_lock:
            try:
                del self.locks[hash_key]
            except KeyError:
                pass
            return True



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
"""Package setup."""
from __future__ import absolute_import, print_function

import io
import re
from glob import glob
from os.path import basename, dirname, join, splitext

from setuptools import find_packages, setup


def read(*names, **kwargs):
    """Read file contents."""
    return io.open(join(dirname(__file__), *names), encoding=kwargs.get("encoding", "utf8")).read()


setup(
    name="tuco",
    version="0.3.0",
    license="MIT license",
    description="A simple class based finite state machine with parsing time validation",
    long_description="%s\n%s"
    % (
        re.compile("^.. start-badges.*^.. end-badges", re.M | re.S).sub("", read("README.rst")),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    author="EatFirst",
    author_email="it+opensource@eatfirst.com",
    url="https://github.com/eatfirst/python-tuco",
    packages=find_packages("src"),
    package_dir={"": "src"},
    py_modules=[splitext(basename(path))[0] for path in glob("src/*.py")],
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Utilities",
    ],
    keywords=[
        # eg: 'keyword1', 'keyword2', 'keyword3',
    ],
    install_requires=["pytz", "typing"],
    extras_require={"graph": ["graphviz >= 0.8.1"], "redis": ["redis >= 2.10"]},
)



"""Graph builder module."""
import os
import tempfile
from typing import List, Optional, Type

from tuco import FSM
from tuco.exceptions import TucoEmptyFSMError
from tuco.properties import Error, Event, FinalState, State, Timeout


class GraphBuilder:
    """Parse state machine's states and events to generate graphs."""

    TIMEOUT_ATTRIBUTES = {"color": "lightgray", "fontcolor": "lightgray"}
    FINAL_STATE_ATTRIBUTES = {"shape": "box", "fillcolor": "lightgray", "style": "filled"}
    EVENT_ATTRIBUTES = {"color": "green", "fontcolor": "green"}
    ERROR_ATTRIBUTES = {"color": "red", "fontcolor": "red"}

    def __init__(self, fsm_class: Type[FSM], file_format) -> None:
        try:
            from graphviz import Digraph
        except ImportError as e:
            raise RuntimeError(
                "Graphviz could not be found, make sure you installed tuco with " "optional graph support."
            ) from e
        states = fsm_class.get_all_states()
        if not states:
            raise TucoEmptyFSMError()

        self.fsm_class = fsm_class
        self.states = states.items()
        self.file_format = file_format

        comment = self.fsm_class.__doc__ or ""
        # Replace line endings for multi line comments.
        comment = comment.replace("\n", "\n//")

        self.dot = Digraph(comment=comment, format=file_format, graph_attr={"overlap": "false"})

    @classmethod
    def generate_from_class(cls, fsm_class: Type[FSM], file_format) -> str:
        """Generate a SVG graph from a tuco fsm class."""
        generation = cls(fsm_class, file_format)
        generation.iterate_states()
        return generation.render_graph()

    def iterate_states(self) -> None:
        """Iterate over all states sent and add nodes to graphviz."""
        for state_name, state in self.states:
            if isinstance(state, FinalState):
                self.add_final_state_node(state_name)
            else:
                self.add_node(state, state_name)

    def iterate_events(self, parent_state_name: str, events: List[Event]) -> None:
        """Iterate over all events of a state and add nodes to graphviz."""
        for event in events:
            self.dot.node(event.target_state)
            self.dot.edge(
                parent_state_name,
                event.target_state,
                label="Event: {}".format(event.event_name),
                _attributes=self.EVENT_ATTRIBUTES,
            )

            self.add_error_node(parent_state_name, event.error)

    def add_node(self, state: State, state_name: str) -> None:
        """Add a node with its events and timeouts."""
        self.dot.node(state_name)

        self.add_error_node(state_name, state.error)
        self.add_timeout_node(state_name, state.timeout)

        self.iterate_events(state_name, state.events)

    def add_final_state_node(self, state_name: str) -> None:
        """Add a final node."""
        self.dot.node(state_name, _attributes=self.FINAL_STATE_ATTRIBUTES)

    def add_error_node(self, parent_state_name: str, error: Optional[Error]) -> None:
        """Render an error node if present."""
        if not error:
            return

        target_state = error.target_state
        self.dot.node(target_state)
        self.dot.edge(parent_state_name, target_state, label="Error", _attributes=self.ERROR_ATTRIBUTES)

    def add_timeout_node(self, parent_state_name: str, timeout: Optional[Timeout]) -> None:
        """Render a timeout node if present."""
        if not timeout:
            return
        self.dot.node(timeout.target_state)
        self.dot.edge(parent_state_name, timeout.target_state, label="Timeout", _attributes=self.TIMEOUT_ATTRIBUTES)

    def render_graph(self) -> str:
        """Render graph in a file and return its content."""
        temp_file = tempfile.NamedTemporaryFile(prefix="tuco", delete=False)
        temp_file.close()
        filename = temp_file.name
        graph_file = "{}.{}".format(filename, self.file_format)

        try:
            self.dot.render(os.path.basename(filename), os.path.dirname(filename))
            return open(graph_file).read()
        finally:
            for name in (filename, graph_file):
                try:
                    os.unlink(name)
                except OSError:
                    pass


generate_from_class = GraphBuilder.generate_from_class
